using System;
using System.Runtime.InteropServices;
using System.Windows.Forms;

#region Clarion basic datatypes that exactly match to .NET datatypes
using CWByte     = System.Byte;
using CWShort    = System.Int16;
using CWUShort   = System.UInt16;
using CWLong     = System.Int32;
using CWULong    = System.UInt32;
using CWSigned   = System.Int32;
using CWUnsigned = System.UInt32;
using CWSReal    = System.Single;
using CWReal     = System.Double;
#endregion

namespace CWInterop
{
   public class CWLink
   {
      #region External Interface for SharedDLL04 Clarion DLL
      [MarshalAs(UnmanagedType.BStr)]
      static Type Zx = typeof(System.String);

      [DllImport("SharedDLL04", EntryPoint="PUBLICPROCEDURE")]
      public static extern void PublicProcedure();

      [DllImport("SharedDLL04", EntryPoint="TAKESLONGRETURNSLONG")]
      public static extern CWLong TakesLongReturnsLong(CWLong value);

      [DllImport("SharedDLL04", EntryPoint="TAKESSTRING")]
      public static extern void TakesString([MarshalAs(UnmanagedType.BStr)] string value);

      [DllImport("SharedDLL04", EntryPoint="TAKESANDRETURNSSTRING")]
      [return: MarshalAs(CWBString)]
      public static extern string TakesAndReturnsString([MarshalAs(CWBString)] string value);

      [DllImport("SharedDLL04", EntryPoint="TAKESANDRETURNSDATE")]
      public static extern CWDateLong TakesAndReturnsDate_External(CWDateLong value);
      public static DateTime TakesAndReturnsDate(DateTime value)
      {
         return (DateTime) TakesAndReturnsDate_External(value);
      }

      [DllImport("SharedDLL04", EntryPoint="TAKESANDRETURNSTIME")]
      public static extern CWTimeLong TakesAndReturnsTime_External(CWTimeLong value);
      public static DateTime TakesAndReturnsTime(DateTime value)
      {
         return (DateTime) TakesAndReturnsTime_External(value);
      }

      [StructLayout(LayoutKind.Sequential, Pack=1, CharSet=CharSet.Ansi)]
      public struct TestGroup
      {
         public CWByte  ByteVal;
         public CWShort ShortVal;
         public CWLong  LongVal;
         public CWReal  RealVal;
      }

      [DllImport("SharedDLL04", EntryPoint="TAKESANDRETURNSGROUP")]
      public static extern void TakesAndReturnsGroup(ref TestGroup pTestGroup);

      #endregion

      #region Clarion Datatype Converters
      // Shortcut for Clarion BSTRING to save typing
      const UnmanagedType CWBString = UnmanagedType.BStr;

      /// <summary>
      ///    This structure converts back and forth between a .NET DateTime object and a Clarion DATE.  For
      ///    details see http://www.clarionmag.com.  This uses implicit conversion to convert the DATE portion
      ///    of a DateTime to the 32 bit integer equivalent used by Clarion.   This class struct should
      ///    never be created explicitly.  It only exists to marshal data betwen Clarion and .NET.
      /// </summary>
      public struct CWDateLong
      {
         // Date portion of DateTime as a Clarion LONG
         public Int32 _cwDateAsLong;

         // Offset in ticks between .NET start date and Clarion start date.
         const long DATEOFFSET = 657432;

         // Implicit operator to convert the Int32 returned by a Clarion function to a DateTime object.
         public static implicit operator DateTime(CWDateLong obj)
         {
            return new DateTime((obj._cwDateAsLong + DATEOFFSET) * TimeSpan.TicksPerDay);
         }

         // Implicit operator to convert a DateTime to the 32 bit Clarion value
         public static implicit operator CWDateLong(DateTime value)
         {
            CWDateLong result = new CWDateLong();
            result._cwDateAsLong = Convert.ToInt32((value.Ticks / TimeSpan.TicksPerDay) - DATEOFFSET);
            return result;
         }
      }

      /// <summary>
      ///    This structure converts back and forth between a .NET DateTime object and a Clarion TIME.  For
      ///    details see http://www.clarionmag.com.  This uses implicit conversion to convert the TIME portion
      ///    of a DateTime to the 32 bit integer equivalent used by Clarion.   This class struct should
      ///    never be created explicitly.  It only exists to marshal data betwen Clarion and .NET.
      /// </summary>
      public struct CWTimeLong
      {
         // Time portion of DateTime as a Clarion LONG
         private Int32 _cwTimeAsLong;

         // Implicit operator to convert the Int32 returned by a Clarion function to a DateTime object.
         public static implicit operator DateTime(CWTimeLong obj)
         {
            return new DateTime() + new TimeSpan((obj._cwTimeAsLong - 1) * TimeSpan.TicksPerSecond / 100);
         }

         // Implicit operator to convert a DateTime to the 32 bit Clarion value
         public static implicit operator CWTimeLong(DateTime value)
         {
            CWTimeLong result = new CWTimeLong();
            result._cwTimeAsLong = Convert.ToInt32(value.Millisecond / 10 + value.Second * 100
                                 + value.Minute * 6000 + value.Hour * 360000 + 1);
            return result;
         }
      }
      #endregion
   }
}
